"use strict";
/*!
 * Copyright 2020-2024 Google LLC
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 *      http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */
Object.defineProperty(exports, "__esModule", { value: true });
exports.legacyExports = exports.extractSpan = exports.containsSpanContext = exports.injectSpan = exports.PubsubEvents = exports.PubsubSpans = exports.getTopicInfo = exports.getSubscriptionInfo = exports.legacyAttributeName = exports.modernAttributeName = exports.spanContextToContext = exports.pubsubSetter = exports.pubsubGetter = exports.PubsubMessageSet = exports.PubsubMessageGet = exports.PubsubMessageGetSet = exports.isEnabled = exports.OpenTelemetryLevel = void 0;
const api_1 = require("@opentelemetry/api");
// We need this to get the library version.
// eslint-disable-next-line @typescript-eslint/no-var-requires
const packageJson = require('../../package.json');
/**
 * Instantiates a Opentelemetry tracer for the library
 *
 * @private
 */
let cachedTracer;
function getTracer() {
    const tracer = cachedTracer !== null && cachedTracer !== void 0 ? cachedTracer : api_1.trace.getTracer('@google-cloud/pubsub', packageJson.version);
    cachedTracer = tracer;
    return cachedTracer;
}
/**
 * Determination of the level of OTel support we're providing.
 *
 * @private
 */
var OpenTelemetryLevel;
(function (OpenTelemetryLevel) {
    /**
     * None: OTel support is not enabled because we found no trace provider.
     */
    OpenTelemetryLevel[OpenTelemetryLevel["None"] = 0] = "None";
    /**
     * Legacy: We found a trace provider, but the user also specified the old
     * manual enable flag; this will trigger the legacy attribute being included.
     * The modern propagation attribute will _also_ be included.
     */
    OpenTelemetryLevel[OpenTelemetryLevel["Legacy"] = 1] = "Legacy";
    /**
     * Modern: We will only inject/extract the modern propagation attribute.
     */
    OpenTelemetryLevel[OpenTelemetryLevel["Modern"] = 2] = "Modern";
})(OpenTelemetryLevel || (exports.OpenTelemetryLevel = OpenTelemetryLevel = {}));
/**
 * Tries to divine what sort of OpenTelemetry we're supporting. See the enum
 * for the meaning of the values, and other notes.
 *
 * Legacy OTel is no longer officially supported, but we don't want to
 * break anyone at a non-major.
 *
 * @private
 */
function isEnabled(publishSettings) {
    // If there's no trace provider attached, do nothing in any case.
    const traceProvider = api_1.trace.getTracerProvider();
    if (!traceProvider) {
        return OpenTelemetryLevel.None;
    }
    if (publishSettings === null || publishSettings === void 0 ? void 0 : publishSettings.enableOpenTelemetryTracing) {
        return OpenTelemetryLevel.Legacy;
    }
    return OpenTelemetryLevel.Modern;
}
exports.isEnabled = isEnabled;
/**
 * Implements common members for the TextMap getter and setter interfaces for Pub/Sub messages.
 *
 * @private
 */
class PubsubMessageGetSet {
    keys(carrier) {
        return Object.getOwnPropertyNames(carrier.attributes)
            .filter(n => n.startsWith(PubsubMessageGetSet.keyPrefix))
            .map(n => n.substring(PubsubMessageGetSet.keyPrefix.length));
    }
    attributeName(key) {
        return `${PubsubMessageGetSet.keyPrefix}${key}`;
    }
}
exports.PubsubMessageGetSet = PubsubMessageGetSet;
PubsubMessageGetSet.keyPrefix = 'googclient_';
/**
 * Implements the TextMap getter interface for Pub/Sub messages.
 *
 * @private
 */
class PubsubMessageGet extends PubsubMessageGetSet {
    get(carrier, key) {
        var _a;
        return (_a = carrier === null || carrier === void 0 ? void 0 : carrier.attributes) === null || _a === void 0 ? void 0 : _a[this.attributeName(key)];
    }
}
exports.PubsubMessageGet = PubsubMessageGet;
/**
 * Implements the TextMap setter interface for Pub/Sub messages.
 *
 * @private
 */
class PubsubMessageSet extends PubsubMessageGetSet {
    set(carrier, key, value) {
        if (!carrier.attributes) {
            carrier.attributes = {};
        }
        carrier.attributes[this.attributeName(key)] = value;
    }
}
exports.PubsubMessageSet = PubsubMessageSet;
/**
 * The getter to use when calling extract() on a Pub/Sub message.
 *
 * @private
 */
exports.pubsubGetter = new PubsubMessageGet();
/**
 * The setter to use when calling inject() on a Pub/Sub message.
 *
 * @private
 */
exports.pubsubSetter = new PubsubMessageSet();
/**
 * Converts a SpanContext to a full Context, as needed.
 *
 * @private
 */
function spanContextToContext(parent) {
    return parent ? api_1.trace.setSpanContext(api_1.context.active(), parent) : undefined;
}
exports.spanContextToContext = spanContextToContext;
/**
 * The modern propagation attribute name.
 *
 * Technically this is determined by the OpenTelemetry library, but
 * in practice, it follows the W3C spec, so this should be the right
 * one. The only thing we're using it for, anyway, is emptying user
 * supplied attributes.
 *
 * @private
 */
exports.modernAttributeName = 'googclient_traceparent';
/**
 * The old legacy attribute name.
 *
 * @private
 */
exports.legacyAttributeName = 'googclient_OpenTelemetrySpanContext';
/**
 * Break down the subscription's full name into its project and ID.
 *
 * @private
 */
function getSubscriptionInfo(fullName) {
    const results = fullName.match(/projects\/([^/]+)\/subscriptions\/(.+)/);
    if (!(results === null || results === void 0 ? void 0 : results[0])) {
        return {
            subName: fullName,
        };
    }
    return {
        subName: fullName,
        projectId: results[1],
        subId: results[2],
    };
}
exports.getSubscriptionInfo = getSubscriptionInfo;
/**
 * Break down the subscription's full name into its project and ID.
 *
 * @private
 */
function getTopicInfo(fullName) {
    const results = fullName.match(/projects\/([^/]+)\/topics\/(.+)/);
    if (!(results === null || results === void 0 ? void 0 : results[0])) {
        return {
            topicName: fullName,
        };
    }
    return {
        topicName: fullName,
        projectId: results[1],
        topicId: results[2],
    };
}
exports.getTopicInfo = getTopicInfo;
class PubsubSpans {
    static createAttributes(params, message) {
        var _a, _b, _c, _d;
        const destinationName = (_a = params.topicName) !== null && _a !== void 0 ? _a : params.subName;
        const destinationId = (_b = params.topicId) !== null && _b !== void 0 ? _b : params.subId;
        const projectId = params.projectId;
        // Purposefully leaving this debug check here as a comment - this should
        // always be true, but we don't want to fail in prod if it's not.
        /*if (
          (params.topicName && params.subName) ||
          (!destinationName && !projectId && !destinationId)
        ) {
          throw new Error(
            'One of topicName or subName must be specified, and must be fully qualified'
          );
        }*/
        const spanAttributes = {
            // Add Opentelemetry semantic convention attributes to the span, based on:
            // https://github.com/open-telemetry/opentelemetry-specification/blob/v1.1.0/specification/trace/semantic_conventions/messaging.md
            ['messaging.system']: 'gcp_pubsub',
            ['messaging.destination.name']: destinationId !== null && destinationId !== void 0 ? destinationId : destinationName,
            ['gcp.project_id']: projectId,
        };
        if (message) {
            if (message.calculatedSize) {
                spanAttributes['messaging.message.envelope.size'] =
                    message.calculatedSize;
            }
            else {
                if ((_c = message.data) === null || _c === void 0 ? void 0 : _c.length) {
                    spanAttributes['messaging.message.envelope.size'] =
                        (_d = message.data) === null || _d === void 0 ? void 0 : _d.length;
                }
            }
            if (message.orderingKey) {
                spanAttributes['messaging.gcp_pubsub.message.ordering_key'] =
                    message.orderingKey;
            }
            if (message.isExactlyOnceDelivery) {
                spanAttributes['messaging.gcp_pubsub.message.exactly_once_delivery'] =
                    message.isExactlyOnceDelivery;
            }
            if (message.ackId) {
                spanAttributes['messaging.gcp_pubsub.message.ack_id'] = message.ackId;
            }
        }
        return spanAttributes;
    }
    static createPublisherSpan(message, topicName) {
        const topicInfo = getTopicInfo(topicName);
        const span = getTracer().startSpan(`${topicName} create`, {
            kind: api_1.SpanKind.PRODUCER,
            attributes: PubsubSpans.createAttributes(topicInfo, message),
        });
        if (topicInfo.topicId) {
            span.updateName(`${topicInfo.topicId} create`);
            span.setAttribute('messaging.destination.name', topicInfo.topicId);
        }
        return span;
    }
    static updatePublisherTopicName(span, topicName) {
        const topicInfo = getTopicInfo(topicName);
        if (topicInfo.topicId) {
            span.updateName(`${topicInfo.topicId} create`);
            span.setAttribute('messaging.destination.name', topicInfo.topicId);
        }
        else {
            span.updateName(`${topicName} create`);
        }
        if (topicInfo.projectId) {
            span.setAttribute('gcp.project_id', topicInfo.projectId);
        }
    }
    static createReceiveSpan(message, subName, parent) {
        var _a;
        const subInfo = getSubscriptionInfo(subName);
        const name = `${(_a = subInfo.subId) !== null && _a !== void 0 ? _a : subName} subscribe`;
        const attributes = this.createAttributes(subInfo, message);
        if (subInfo.subId) {
            attributes['messaging.destination.name'] = subInfo.subId;
        }
        if (api_1.context) {
            return getTracer().startSpan(name, {
                kind: api_1.SpanKind.CONSUMER,
                attributes,
            }, parent);
        }
        else {
            return getTracer().startSpan(name, {
                kind: api_1.SpanKind.CONSUMER,
                attributes,
            });
        }
    }
    static createChildSpan(name, message, parentSpan, attributes) {
        var _a;
        const parent = (_a = message === null || message === void 0 ? void 0 : message.parentSpan) !== null && _a !== void 0 ? _a : parentSpan;
        if (parent) {
            return getTracer().startSpan(name, {
                kind: api_1.SpanKind.INTERNAL,
                attributes: attributes !== null && attributes !== void 0 ? attributes : {},
            }, spanContextToContext(parent.spanContext()));
        }
        else {
            return undefined;
        }
    }
    static createPublishFlowSpan(message) {
        return PubsubSpans.createChildSpan('publisher flow control', message);
    }
    static createPublishSchedulerSpan(message) {
        return PubsubSpans.createChildSpan('publisher batching', message);
    }
    static createPublishRpcSpan(messages, topicName) {
        const spanAttributes = PubsubSpans.createAttributes(getTopicInfo(topicName));
        const links = messages
            .map(m => { var _a; return ({ context: (_a = m.parentSpan) === null || _a === void 0 ? void 0 : _a.spanContext() }); })
            .filter(l => l.context);
        const span = getTracer().startSpan(`${topicName} send`, {
            kind: api_1.SpanKind.PRODUCER,
            attributes: spanAttributes,
            links,
        }, api_1.ROOT_CONTEXT);
        span === null || span === void 0 ? void 0 : span.setAttribute('messaging.batch.message_count', messages.length);
        messages.forEach(m => {
            // Workaround until the JS API properly supports adding links later.
            if (m.parentSpan) {
                m.parentSpan.setAttribute('messaging.gcp_pubsub.publish.trace_id', span.spanContext().traceId);
                m.parentSpan.setAttribute('messaging.gcp_pubsub.publish.span_id', span.spanContext().spanId);
            }
        });
        return span;
    }
    static createReceiveResponseRpcSpan(messageSpans, subName) {
        const spanAttributes = PubsubSpans.createAttributes(getSubscriptionInfo(subName));
        const links = messageSpans
            .map(m => ({ context: m === null || m === void 0 ? void 0 : m.spanContext() }))
            .filter(l => l.context);
        const span = getTracer().startSpan(`${subName} response batch`, {
            kind: api_1.SpanKind.CONSUMER,
            attributes: spanAttributes,
            links,
        }, api_1.ROOT_CONTEXT);
        span === null || span === void 0 ? void 0 : span.setAttribute('messaging.batch.message_count', messageSpans.length);
        messageSpans.forEach(m => {
            // Workaround until the JS API properly supports adding links later.
            if (m) {
                m.setAttribute('messaging.gcp_pubsub.receive.trace_id', span.spanContext().traceId);
                m.setAttribute('messaging.gcp_pubsub.receive.span_id', span.spanContext().spanId);
            }
        });
        return span;
    }
    static createReceiveFlowSpan(message) {
        return PubsubSpans.createChildSpan('subscriber concurrency control', message);
    }
    static createReceiveSchedulerSpan(message) {
        return PubsubSpans.createChildSpan('subscriber scheduler', message);
    }
    static createReceiveProcessSpan(message, subName) {
        var _a;
        const subInfo = getSubscriptionInfo(subName);
        return PubsubSpans.createChildSpan(`${(_a = subInfo.subId) !== null && _a !== void 0 ? _a : subName} process`, message);
    }
    static setReceiveProcessResult(span, isAck) {
        span.setAttribute('messaging.gcp_pubsub.result', isAck ? 'ack' : 'nack');
    }
    static createReceiveLeaseSpan(message, subName, type, deadline, isInitial) {
        var _a;
        const subInfo = getSubscriptionInfo(subName);
        const span = PubsubSpans.createReceiveSpan(message, `${(_a = subInfo.subId) !== null && _a !== void 0 ? _a : subInfo.subName} ${type}`, undefined);
        if (deadline) {
            span === null || span === void 0 ? void 0 : span.setAttribute('messaging.gcp_pubsub.message.ack_deadline_seconds', deadline.totalOf('second'));
        }
        if (isInitial !== undefined) {
            span === null || span === void 0 ? void 0 : span.setAttribute('messaging.gcp_pubsub.is_receipt_modack', isInitial);
        }
        return span;
    }
}
exports.PubsubSpans = PubsubSpans;
/**
 * Creates and manipulates Pub/Sub-related events on spans.
 */
class PubsubEvents {
    static addEvent(text, message, attributes) {
        const parent = message.parentSpan;
        if (!parent) {
            return;
        }
        parent.addEvent(text, attributes);
    }
    static publishStart(message) {
        PubsubEvents.addEvent('publish start', message);
    }
    static publishEnd(message) {
        PubsubEvents.addEvent('publish end', message);
    }
    static ackStart(message) {
        PubsubEvents.addEvent('ack start', message);
    }
    static ackEnd(message) {
        PubsubEvents.addEvent('ack end', message);
    }
    static modackStart(message) {
        PubsubEvents.addEvent('modack start', message);
    }
    static modackEnd(message) {
        PubsubEvents.addEvent('modack end', message);
    }
    static nackStart(message) {
        PubsubEvents.addEvent('nack start', message);
    }
    static nackEnd(message) {
        PubsubEvents.addEvent('nack end', message);
    }
    static ackCalled(span) {
        span.addEvent('ack called');
    }
    static nackCalled(span) {
        span.addEvent('nack called');
    }
    static modAckCalled(span, deadline) {
        // User-called modAcks are never initial ones.
        span.addEvent('modack called', {
            'messaging.gcp_pubsub.modack_deadline_seconds': `${deadline.totalOf('second')}`,
            'messaging.gcp_pubsub.is_receipt_modack': 'false',
        });
    }
    static modAckStart(message, deadline, isInitial) {
        PubsubEvents.addEvent('modack start', message, {
            'messaging.gcp_pubsub.modack_deadline_seconds': `${deadline.totalOf('second')}`,
            'messaging.gcp_pubsub.is_receipt_modack': isInitial ? 'true' : 'false',
        });
    }
    static modAckEnd(message) {
        PubsubEvents.addEvent('modack end', message);
    }
    // Add this event any time the process is shut down before processing
    // of the message can complete.
    static shutdown(message) {
        PubsubEvents.addEvent('shutdown', message);
    }
}
exports.PubsubEvents = PubsubEvents;
/**
 * Injects the trace context into a Pub/Sub message (or other object with
 * an 'attributes' object) for propagation.
 *
 * This is for the publish side.
 *
 * @private
 */
function injectSpan(span, message, enabled) {
    if (!message.attributes) {
        message.attributes = {};
    }
    if (message.attributes[exports.modernAttributeName]) {
        console.warn(`${exports.modernAttributeName} key set as message attribute, but will be overridden.`);
        delete message.attributes[exports.modernAttributeName];
    }
    // If we're in legacy mode, add that header as well.
    if (enabled === OpenTelemetryLevel.Legacy) {
        if (message.attributes[exports.legacyAttributeName]) {
            console.warn(`${exports.legacyAttributeName} key set as message attribute, but will be overridden.`);
        }
        message.attributes[exports.legacyAttributeName] = JSON.stringify(span.spanContext());
    }
    // Always do propagation injection with the trace context.
    const context = api_1.trace.setSpanContext(api_1.ROOT_CONTEXT, span.spanContext());
    api_1.propagation.inject(context, message, exports.pubsubSetter);
    // Also put the direct reference to the Span object for while we're
    // passing it around in the client library.
    message.parentSpan = span;
}
exports.injectSpan = injectSpan;
/**
 * Returns true if this message potentially contains a span context.
 *
 * @private
 */
function containsSpanContext(message) {
    if (message.parentSpan) {
        return true;
    }
    if (!message.attributes) {
        return false;
    }
    const keys = Object.getOwnPropertyNames(message.attributes);
    return !!keys.find(n => n === exports.legacyAttributeName || n === exports.modernAttributeName);
}
exports.containsSpanContext = containsSpanContext;
/**
 * Extracts the trace context from a Pub/Sub message (or other object with
 * an 'attributes' object) from a propagation, for receive processing. If no
 * context was present, create a new parent span.
 *
 * This is for the receive side.
 *
 * @private
 */
function extractSpan(message, subName, enabled) {
    var _a, _b;
    if (message.parentSpan) {
        return message.parentSpan;
    }
    const keys = Object.getOwnPropertyNames((_a = message.attributes) !== null && _a !== void 0 ? _a : {});
    let context;
    if (enabled === OpenTelemetryLevel.Legacy) {
        // Only prefer the legacy attributes to no trace context attribute.
        if (keys.includes(exports.legacyAttributeName) &&
            !keys.includes(exports.modernAttributeName)) {
            const legacyValue = (_b = message.attributes) === null || _b === void 0 ? void 0 : _b[exports.legacyAttributeName];
            if (legacyValue) {
                const parentSpanContext = legacyValue
                    ? JSON.parse(legacyValue)
                    : undefined;
                if (parentSpanContext) {
                    context = spanContextToContext(parentSpanContext);
                }
            }
        }
    }
    else {
        if (keys.includes(exports.modernAttributeName)) {
            context = api_1.propagation.extract(api_1.ROOT_CONTEXT, message, exports.pubsubGetter);
        }
    }
    const span = PubsubSpans.createReceiveSpan(message, subName, context);
    message.parentSpan = span;
    return span;
}
exports.extractSpan = extractSpan;
// Since these were exported on the main Pub/Sub index in the previous
// version, we have to export them until the next major.
exports.legacyExports = {
    /**
     * @deprecated
     * Use the new telemetry functionality instead; see the updated OpenTelemetry
     * sample for an example.
     */
    createSpan: function (spanName, kind, attributes, parent) {
        return getTracer().startSpan(spanName, {
            kind,
            attributes,
        }, parent ? api_1.trace.setSpanContext(api_1.context.active(), parent) : undefined);
    },
};
//# sourceMappingURL=telemetry-tracing.js.map